/**
* HTMLBuilder
*
* A micro-toolkit designed to make creating/manipulating DOM Elements
* fast and easy.
*
* This file contains the templetor component.
*
* Copyright 2010, Carlos Killpack
* Dual licensed under New BSD and WTFPL licenses
*/
if (!(typeof String.prototype.trim === 'function')) {
    String.prototype.trim = function () {
        var self = this;
        return self.replace(/^\s\s*/, '').replace(/\s\s*$/, '');
    };
}
if (!(typeof Array.isArray === 'function')) {
    Array.isArray = function (o) {
        return Object.prototype.toString.call(o) === '[object Array]';
    };
}
if (!(typeof Array.prototype.forEach === 'function')) {
    Array.prototype.forEach = function (fn) {
        var self = this;
        for (var i = 0, l = self.length; i < l; i += 1) {
            fn(self[i]);
        }
    };
}
(function (window) {
    var templates = {},
    newID = function () {
        return String(Math.floor(Math.random() * 100000000001));
    },
    metaVar = function (valueSet, elementExpression, simpleVar) {
    	// Can this be made more readable?
    	return (simpleVar === '') ?
    		'%[' + [[valueSet, elementExpression.join('>')].join('='), simpleVar].join('::') + ']%' :
    		'%[' + [valueSet, elementExpression.join('>')].join('=') + ']%';
    },
    /*
     * Recursively extract tokens from the template and expand
     * variables, either simple or complex, until the template
     * reaches a renderable state.
     */
    process = function (template, data) {
        var rendered = [],
            // Regex for simple variable
            re = /^%\s*(.*?)\s*%$/,
            // Regex for meta variable
            xre = /^\[\s+(.*?)=(.*?)(?:::(.*?)|\s+)\]$/;
        template.forEach(function $process_element (el) {
            /*
             * (1a)
             * If the current element is an Array send it to
             * process() to make an attempt at rendering it,
             * otherwise...
             */
            if (Array.isArray(el)) {
                setTimeout(function () {
                    rendered.push(process(el, data));
                }, 0);
            /*
             * (1b)
             * Check if the current element is a string, if so
             * get ready to do the real work, otherwise...
             */
            } else if (typeof el === 'string') {
                /*
                 * (2a)
                 * Check for a match to the Simple Variable
                 * regex. If it matches process the variable,
                 * otherwise...
                 */ // Trim the string to avoid false negatives
                if (el.trim().match(re)) {
                    var tvar = el.trim().match(re)[1];
                    /*
                     * (3a)
                     * Check for a match to the Meta Variable
                     * regex. If it matches process the variable
                     * as a Meta Variable, otherwise...
                     */
                    if (tvar.match(xre)) {
                        var xrem = tvar.match(xre),
                            xvar = xrem[1] || '',
                            xtag = xrem[2].split('>') || '',
                            xattrs = xrem[3] || '';
                        /*
                         * (4a)
                         * Check for a simple tag expression
                         * then build and render an array
                         * containing the tag name, its
                         * attribute data, and its contents,
                         * Otherwise...
                         */
                        if (xtag.length === 1) {
                            if (Array.isArray(data[xvar])) {
                                data[xvar].forEach(function $process_meta_var (d) {
                                    setTimeout(function () {
                                        rendered.push(
                                        	(xattrs === '') ?
                                        		process([xtag[0], d], data) :
                                        		process([xtag[0], xattrs, d], data)
                                        );
                                    }, 0);
                                });
                            }
                        /*
                         * (4b)
                         * Expand complex element expressions
                         * into multiple simple element
                         * expressions and render them.
                         */
                        } else {
                            var xch = xtag.splice(0, 1)[0];
                            /*
                             * (5a)
                             * If data[xvar] looks like a list
                             * of lists take each list and save 
                             * it with a unique name and create/
                             * parse a meta variable using it as
                             * the value set for the variable,
                             * otherwise...
                             */
                            if (Array.isArray(data[xvar]) && Array.isArray(data[xvar][0])) {
                                data[xvar].forEach(function (li) {
                                    var id = newID();
                                    data[id] = li;
                                    setTimeout(function () {
                                        rendered.push(
                                        	Array.isArray(li) ?
			                                    process([xch, metaVar(id, xtag, xattrs)], data) :
			                                    li
                                        );
                                    }, 0);
                                });
                            /*
                             * (5b)
                             * Create/parse a meta variable with
                             * a simplified element expression
                             */
                            } else {
                                setTimeout(function () {
                                    rendered.push(
                                    	process([xch, metaVar(xvar, xtag, xattrs)], data)
                                    );
                                }, 0);
                            }
                        }
                    /*
                     * (3b)
                     * Do a simple subsititution and push into
                     * the "rendered" array.
                     */
                    } else {
                        setTimeout(function () {
                            data[tvar] && rendered.push(data[tvar]);
                        }, 0);
                    }
                /*
                 * (2b)
                 * Since this isn't a variable there isn't
                 * anything left for us to do; push it into the
                 * "rendered" array.
                 */
                } else {
                    setTimeout(function () {
                        rendered.push(el);
                    }, 0);
                }
            /*
             * (1c)
             * If the current element doesn't match anything
             * that needs to be dealt with push it into the
             * "rendered" array.
             */
            } else {
                setTimeout(function () {
                    rendered.push(el);
                }, 0);
            }
        });
        return rendered;
    };
    /**
    * @constructor
    * Template constructor. Returns an instance of the Templetor
    * object if it's called the name of a predefined template.
    *
    * If it's called with a template definition it stores the template
    * (allowing you to use it later) and returns an instance of the
    * Templetor object.
    *
    * A template definition takes the following form:
    *
    * {
    * 		"name": <templateName>,
    * 		"def": {
    *     		<sectionName>: [ (HTMLBuilder elDef)
    *         		<tagname>,
    *         		{<key>: <value>}
    *         		"% <var> %" || <content> || [...],
    *         		"%<var>%" || <content> || [...],
    *         		...
    *     		],
    *     		<sectionName>: [...],
    *     		"__wrapper__": <elName> || [elDef], (Default: fragment)
    *     		"__sections__": [<sectionName>, ...] (in order)
    * 		}
    * }
    *
    */
    var Templetor = function (arg) { // Wishing for **kwargs..
        var self = this, template;
        if (typeof arg === 'string') {
            template = templates[arg];
        } else if (!Array.isArray(arg) && typeof arg === 'object') {
            templates[arg.name] = arg.def;
            template = arg.def;
        } else {
            throw 'Invalid argument: unexpected string or object.';
        }
        self.rendered = false;
        self.template = (function (templateDef) {
            var renderable = [];
            if (templateDef.__wrapper__) {
                renderable = templateDef.__wrapper__ instanceof Array ?
                    templateDef.__wrapper__ : [templateDef.__wrapper__];
            }
            templateDef.__sections__.forEach(function $process_raw (sec) {
                renderable.push(templateDef[sec]);
            });
            return renderable;
        }(template));
    };
    Templetor.prototype.version = '0.1';
    /*
    * Takes an object literal that contains properties with the same
    * names as the variables used in the template being built and
    * builds the template recursively.
    */
    Templetor.prototype.compile = function (data, callback) {
        var self = this;
        if (!self.rendered) {
            self.rendered = process(self.template, data);
        }
        return callback ? callback(self.rendered) : self.rendered;
    };
    return (window.Templetor = Templetor);
}(window));
