/**
* HTMLBuilder
*
* A micro-toolkit designed to make creating/manipulating DOM Elements
* fast and easy.
*
* This file contains the wildebeest template engine component.
*
* Copyright 2010, Carlos Killpack
* Dual licensed under New BSD and WTFPL licenses
*/
if (!(typeof Array.prototype.forEach === 'function')) {
    Array.prototype.forEach = function (fn) {
        'use strict';
        for (var i = 0, l = this.length; i < l; i += 1) {
            fn(this[i]);
        }
    };
}
if (!(typeof String.prototype.trim === 'function')) {
    String.prototype.trim = function () {
        return this.replace(/^\s\s*/, '').replace(/\s\s*$/, '');
    };
}
if (!(typeof Object.keys === 'function')) {
    Object.keys = function (o) {
        'use strict';
        var key, keys = [];
        for (key in o) {
            o.hasOwnProperty(key) && keys.push(key);
        }
        return keys;
    };
}
if (!(typeof Array.prototype.map === 'function')) {
    Array.prototype.map = function (fn) {
        'use strict';
        if (this === void 0 || this === null) {
            throw new TypeError();
        }
        var t = Object(this),
            len = t.length >>> 0;
        if (typeof fn !== 'function') {
            throw new TypeError();
        }
        var res = new Array(len),
            thisp = arguments[1];
        for (var i = 0; i < len; i += 1) {
            if (i in t) {
                res[i] = fn.call(thisp, t[i], i, t);
            }
        }
        return res;
    };
}
(function (window) {
    'use strict';
    var _macro = {
            '/list': function (data) {
                return Object.keys(data).map(function $list_map (el) {
                    return '<li>' + data[el] + '</li>';
                }).join('');
            },
            '/dlist': function (data) {
                return Object.keys(data).map(function $dlist_map (dt) {
                    return [
                        '<dt>' + dt + '</dt>',
                        '<dd>' + data[dt] + '</dt>'
                    ].join('');
                }).join('');
            }
        },
        Wildebeest = function (macro) {
            var self = this,
                makeKey = function (tmpl, data) {
                    return Object.keys(data).map(function $mk_k_map (el) {
                        return data[el];
                    }) + tmpl;
                },
                cache = {},
                SPLITTER = /\{(?=\{)|\}(?!\})/g,
                EXPRESSION = /^\{\s?([a-zA-Z0-9_\s\.\/]*)\s?\}$/g,
                DOT = /\./g;
            self.compile = function (tmpl, data, key) {
                var key = key ? key : makeKey(tmpl, data);
                if (cache[key]) {
                    return cache[key];
                } 
                if (!/\W/.test(tmpl)) {
                    return self.compile(
                        document.getElementById(tmpl).innerHTML,
                        data,
                        key
                    );
                }
                return (cache[key] = tmpl.split(SPLITTER).map(function $proc_map (el) {
                    return el.replace(EXPRESSION, function $proc_repl (str, p1) {
                        if (p1.charAt(0) === '/') {
                            var segs = p1.split(' '),
                                theMacro = segs.shift(),
                                argsArray = segs.map(function $a_ar_map (seg) {
                                    return eval(['data[', seg.replace(DOT, '\'][\''), ']'].join('\''));
                                }, self);
                            return macro[theMacro].apply(self, argsArray);
                        } else {
                            return eval(['data[', p1.trim().replace(DOT, '\'][\''), ']'].join('\''));
                        }
                    }) || el;
                }).join(''));
            };
            self.addMacro = function (name, fn) {
                return setTimeout(function () {
                    macro['/' + name] = fn;
                }, 0);
            };
        };
    Wildebeest.prototype.version = '0.2.1';
    return function (window) {
        window.wdb = new Wildebeest(_macro);
        window.Wildebeest = Wildebeest;
    };
}(window)(window));
